/**
 * Indicator-Fish
 *
 * Copyright (C) Dylan McCall
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

using AppIndicator;

const string GETTEXT_PACKAGE = "indicator-fish";

class FortuneDialog : Gtk.Dialog {
	private static const int RESPONSE_PREVIOUS = 1;
	private static const int RESPONSE_NEXT = 2;
	
	private const string fortune_title = _("Wanda the Fish says…");
	private const string fortune_image = "/usr/share/indicator-fish/fortune.png";
	
	private List<string> messages;
	private uint messages_position;
	
	private Gtk.Label message_label;
	
	private Gtk.Widget previous_button;
	private Gtk.Widget next_button;
	private Gtk.Widget close_button;
	
	public signal void new_message();
	
	public FortuneDialog() {
		Object();
		
		this.messages = new List<string>();
		this.messages_position = 0;
		
		this.set_position (Gtk.WindowPosition.CENTER);
		this.set_resizable(false);
		this.set_title (fortune_title);
		this.set_skip_taskbar_hint (true);
		
		this.previous_button = this.add_button(_("_Previous"), RESPONSE_PREVIOUS);
		previous_button.set_sensitive(false);
		this.next_button = this.add_button(_("_Next"), RESPONSE_NEXT);
		//this.next_button.grab_focus();
		this.close_button = this.add_button(Gtk.Stock.CLOSE, Gtk.ResponseType.CLOSE);
		
		this.set_default_response(Gtk.ResponseType.CLOSE);
		
		this.response.connect((response_id) => {
			if (response_id == Gtk.ResponseType.CLOSE) this.destroy();
			if (response_id == RESPONSE_PREVIOUS) this.show_previous();
			if (response_id == RESPONSE_NEXT) this.show_next();
		});
		
		Gtk.Box content = (Gtk.Box) this.get_content_area ();
		
		content.set_margin_top(6);
		content.set_margin_right(6);
		content.set_margin_bottom(6);
		content.set_margin_left(6);
		
		Gtk.ScrolledWindow message_scroll = new Gtk.ScrolledWindow(null, null);
		message_scroll.set_vexpand(true);
		message_scroll.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC);
		message_scroll.set_min_content_height(180);
		content.add (message_scroll);
		
		Gtk.Grid message_box = new Gtk.Grid ();
		message_box.set_hexpand(true);
		message_box.set_halign(Gtk.Align.FILL);
		message_box.set_column_spacing (12);
		message_scroll.add_with_viewport (message_box);
		
		Pango.AttrList message_attributes = new Pango.AttrList ();
		message_attributes.insert (Pango.attr_scale_new (1.1));
		
		this.message_label = new Gtk.Label (null);
		message_label.set_hexpand(true);
		message_label.set_padding (6, 6);
		message_label.set_attributes (message_attributes);
		message_label.set_selectable(true);
		message_label.set_width_chars (72); // fortunes should be no wider than 72 characters
		message_label.set_line_wrap (true);
		message_label.set_line_wrap_mode (Pango.WrapMode.WORD_CHAR);
		message_box.attach (message_label, 0, 0, 1, 1);
		
		Gtk.Image message_image = new Gtk.Image.from_file (fortune_image);
		message_image.set_alignment (1, 0);
		message_box.attach (message_image, 1, 0, 1, 1);
		message_image.set_padding (6, 12);
		
		content.show_all();
	}
	
	private void show_message(uint message_index) {
		string message = this.messages.nth(message_index).data;
		
		this.message_label.set_text(message);
		
		this.messages_position = message_index;
		
		if (this.messages_position == 0) {
			this.previous_button.set_sensitive(false);
		} else {
			this.previous_button.set_sensitive(true);
		}
	}
	
	private void show_previous() {
		if (this.messages_position > 0) {
			this.show_message(this.messages_position - 1);
		}
	}
	
	private void show_next() {
		if (this.messages_position+1 < this.messages.length()) {
			this.show_message(this.messages_position+1);
		} else {
			this.new_message();
		}
	}
	
	public void push_message(string new_message) {
		this.messages.append(new_message);
		this.show_message(this.messages.length()-1);
	}
}

class FishIndicator {
	private Indicator indicator;
	
	private const string icon_format = "/usr/share/indicator-fish/wanda-%d.png";
	private const int icon_count = 8;
	private const int icon_change_delay = 1;
	private int current_icon;
	
	private FortuneDialog fortune_dialog;
	
	private string fortune_program;
	
	public FishIndicator (string fortune_program = "fortune") {
		this.indicator = new Indicator ("indicator-fish",
		                                "",
		                                IndicatorCategory.APPLICATION_STATUS);
		
		indicator.set_title(_("Wanda the Fish"));
		
		this.fortune_program = fortune_program;
		
		this.set_icon(0);
		Timeout.add_seconds (this.icon_change_delay, this.icon_change_timeout);
		
		Gtk.Menu menu = new Gtk.Menu ();
		Gtk.MenuItem something_fishy = new Gtk.MenuItem ();
		something_fishy.set_label (_("Wanda says…"));
		menu.append (something_fishy);
		
		something_fishy.activate.connect (this.something_fishy_clicked);
		
		menu.show_all ();
		
		indicator.set_menu (menu);
		indicator.set_secondary_activate_target (something_fishy);
		
		this.indicator.set_status (IndicatorStatus.ACTIVE);
	}
	
	private void create_fortune_dialog () {
		this.fortune_dialog = new FortuneDialog ();
		
		this.fortune_dialog.new_message.connect (() => {
			this.say_something_fishy();
		});
		
		this.fortune_dialog.destroy.connect (() => {
			this.fortune_dialog = null;
		});
	}
	
	private void show_message (string message) {
		if (this.fortune_dialog == null) {
			this.create_fortune_dialog ();
		}
		
		this.fortune_dialog.push_message (message);
		this.fortune_dialog.present ();
	}
	
	private void say_something_fishy () {
		try {
			Pid pid;
			int std_out_fd;
			
			/* figure out an array of arguments from the given string */
			string[] argv;
			Shell.parse_argv (this.fortune_program, out argv);
			
			Process.spawn_async_with_pipes (null, argv, null,
			                                SpawnFlags.SEARCH_PATH | SpawnFlags.DO_NOT_REAP_CHILD,
			                                null, out pid, null, out std_out_fd);
			
			uint process_timeout = Timeout.add_seconds (3, () => {
				/* this is taking too long; something has probably gone wrong */
				Posix.kill (pid, Posix.SIGKILL);
				return false;
			});
			
			ChildWatch.add (pid, (pid, status) => {
				if (status > 0) {
					/* process exited strangely */
					this.show_message (_("I'm out of fortune cookies."));
				} else {
					/*string message;
					do {
						string new_line = 
					} while (line*/
					FileStream std_out = FileStream.fdopen (std_out_fd, "r");
					uint8 message_out[4096];
					std_out.read (message_out);
					
					string message_out_tidy = (string) message_out;
					this.show_message (message_out_tidy.strip());
				}
				Process.close_pid (pid);
				Source.remove (process_timeout);
			});
			
		} catch (Error e) {
			this.show_message (_("Danger! Danger!\n%s").printf(e.message));
		}
	}
	
	private void something_fishy_clicked () {
		this.say_something_fishy ();
	}
	
	private void set_icon (int icon_number) {
		this.current_icon = icon_number;
		this.indicator.set_icon_full (this.icon_format.printf(this.current_icon),
		                              _("Fishy"));
	}
	
	private bool icon_change_timeout () {
		/* Make Wanda swim */
		int new_icon = 0;
		if (this.current_icon < this.icon_count-1) new_icon = this.current_icon+1;
		this.set_icon(new_icon);
		
		return true;
	}
}

public static int main (string[] args) {
	Intl.bind_textdomain_codeset( GETTEXT_PACKAGE, "UTF-8" );
	Intl.textdomain( GETTEXT_PACKAGE );
	
	Gtk.init (ref args);
	
	string program = string.joinv(" ", args[1:-1]);
	if (program == "") program = "fortune";
	FishIndicator fish_indicator = new FishIndicator (program);
	
	Gtk.main ();
	
	return 0;
}

